/*******************************************************\
 * SAXPY - Scalar Alpha X Plus Y: y[i] = a*x[i] + y[i] *
 * OpenMP 4.0                                          *
 *=====================================================*
 * Center for Computing and Communication (CCC)        *
 * RWTH Aachen University, Germany                     *
 * Author: Frank Robel (robel@rz.rwth-aachen.de)       *
 * Date: October 2013                                  *
\*******************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <omp.h>

#define N 247463936 // 247463936 * 2 * 8 Byte = 3,6875 GByte
#define ANZ 50


// Declaration of helper functions (see below for details)
void printParameter(unsigned int, double);
void saxpy_host    (unsigned int, double, double*, double*);
int  checkResults  (unsigned int, double*,double*);


int main(int argc, char* argv[])
{
  unsigned int i, j;
  int n = N;
  double t1, t2, t3, t4;
  double a = 2.0f;
  double* restrict x;
  double* restrict y;
  double* restrict h_y; // Pointer on CPU (for later comparison)
  x   = (double*) memalign(64, n * sizeof(double));
  y   = (double*) memalign(64, n * sizeof(double));
  h_y = (double*) memalign(64, n * sizeof(double)); 
  if (x==NULL || y==NULL || h_y==NULL)
  {
    printf("Error: Could not allocate memory on the host\n");
    return 1;
  }

  // Initialize x, y
  // Initialize the memory in the same parallel way like you want to 
  // parallelize the calculation in the function saxpy_host() in line 137
  // TODO: add a pragma for parallelization

  for(i=0; i<n; ++i)
  {
    x[i]=i;
    y[i]=h_y[i]=5.0*i-1.0;
  }
  printParameter(n, a);
  
  // Warmup on the Phi
  // TODO 1: use a target pragma to execute the loop on the device
  // Hint  : don't forget to transfer the data between host and phi
  //         Should both vectors be transverd in both directions?
  // TODO 2: add a pragma for parallelization


  for(i=0; i<n; ++i)
  {
    y[i] = a*x[i] + y[i];
  }

  t1 = omp_get_wtime();
  
  // Run SAXPY on the Phi
  // TODO: use a target data pragma to transfer the data between host and phi
  //       Should both vectors be transverd in both directions?

{
  
  for(j = 0; j < ANZ; j++)
  {
    // TODO 1: use a target pragma to execute the loop on the device
    // Hint  : don't forget to "transfer" the data between host and phi
    //         Should both vectors be transverd in both directions?
    // TODO 2: add a pragma for parallelization


    for(i=0; i<n; ++i)
    {
      y[i] = a*x[i] + y[i];
    }
  }

} // end omp target data

  t2 = omp_get_wtime();

  // Warmup on the host
  saxpy_host(n,a,x,h_y);

  t3 = omp_get_wtime();
  
  // Run SAXPY on the host
  // the pragma novector is used to suppress a FUSED LOOP
  // otherwise the compiler would extremely optimize the function calls
#pragma novector
  for(j = 0; j < ANZ; j++)
  {
    saxpy_host(n,a,x,h_y);
  }
  
  t4 = omp_get_wtime();
  
  // Compare CPU and PHI results
  checkResults(n, h_y, y);

  printf("\n");
  printf("Zeitmessung: \n");
  printf("saxpy_phi:\t%f s\n", t2-t1);
  printf("saxpy_host:\t%f s\n", t4-t3);
  printf("\n");

  free(x); free(y); free(h_y);
  return 0;
}

// Print vector size, scalar a and number threads per block
void printParameter(unsigned int n, double a)
{
  printf("Computations are done in double precision\n");
  printf("Vector size n:\t%d\n", n);
  printf("Scalar a:\t%f\n", a);
  printf("Used Memory:\t%5.2f GB\n", (double) n * 2.0 * 8.0 / 1024.0 / 1024.0 / 1024.0);
}

// Compute SAXPY on CPU
void saxpy_host(unsigned int n, double a, double* x, double* y)
{
  unsigned int i;
  // TODO: add a pragma for parallelization

  for(i=0; i<n; ++i)
  {
    y[i] = a*x[i] + y[i];
  }
}

// Compare CPU and GPU results
int checkResults(unsigned int n, double* cpu, double* gpu)
{
  float tolerance = 1e-6f;
  unsigned int i;
  for(i=0; i<n; ++i)
  {
    if (abs(cpu[i] - gpu[i]) > tolerance)
    {
      printf("%d: CPU=%f, GPU=%f\n",i,cpu[i],gpu[i]);
      printf("\nError check failed\n");
      return 1;
    }
  }
  printf("\nError check passed\n");
  
  return 0;
}

